/*!@file
** @brief   The Implementation of the property handler.
** @remarks The XMP Toolkit SDK is used to handle the metadata in PDF files.
** @author  UETA, Shin-ichi <ueta@usdesign.jp>
**
** @copyright (C) 2019 Ueta System Design Studio\n
** This file is licensed under the terms of the MIT License.\n
** See <https://opensource.org/licenses/MIT>.
*/
#pragma once

#include <propsys.h>
#include <propkey.h>
#include <propvarutil.h>
#include <shobjidl.h>

#pragma comment(lib, "propsys.lib")

#include "xmpsdk.h"

/*!
** @brief   The Plain Old Data that specifies a mapping for a property.
** @remarks -
*/
struct CProperty
{
	/// The FMTID/PID identifier that programmatically identifies a property.
	PROPERTYKEY propkey;
	/// The variant type of the property.
	VARTYPE vartype;
	/// The return value from IPropertyStoreCapabilities::IsPropertyWritable.
	HRESULT writable;
	/// The namespace URI for the XMP metadata property to be mapped to.
	XMP_StringPtr schemaNS;
	/// The name of the XMP metadata property to be mapped to.
	XMP_StringPtr propName;
	/// The option bit flags for the XMP metadata property accessor functions.
	XMP_OptionBits options;

	/// The array containing all the values of this object.
	static const CProperty values[];
};

/*!
** @brief   The supplemental option bit flags for the XMP Toolkit.
** @remarks -
*/
enum
{
	kXMP_PropValueIsSimple         = kXMP_NoOptions,

	kXMP_PropValueIsUnorderedArray = kXMP_PropValueIsArray,
	kXMP_PropValueIsOrderedArray   = kXMP_PropValueIsArray          | kXMP_PropArrayIsOrdered,
	kXMP_PropValueIsAlternateArray = kXMP_PropValueIsOrderedArray   | kXMP_PropArrayIsAlternate,
	kXMP_PropValueIsAltTextArray   = kXMP_PropValueIsAlternateArray | kXMP_PropArrayIsAltText,
};

const CProperty
DECLSPEC_SELECTANY
CProperty::values[]{
	{ PKEY_Title,                            VT_LPWSTR,   S_OK, kXMP_NS_DC,  "title",       kXMP_PropValueIsAltTextArray, },
	{ PKEY_Author,               VT_VECTOR | VT_LPWSTR,   S_OK, kXMP_NS_DC,  "creator",     kXMP_PropValueIsOrderedArray, },
	{ PKEY_Subject,                          VT_LPWSTR,   S_OK, kXMP_NS_DC,  "description", kXMP_PropValueIsAltTextArray, },
	{ PKEY_Keywords,             VT_VECTOR | VT_LPWSTR,   S_OK, kXMP_NS_PDF, "Keywords",    kXMP_PropValueIsSimple,       },
	{ PKEY_ApplicationName,                  VT_LPWSTR,   S_OK, kXMP_NS_XMP, "CreatorTool", kXMP_PropValueIsSimple,       },
	{ PKEY_Document_DateCreated,             VT_FILETIME, S_OK, kXMP_NS_XMP, "CreateDate",  kXMP_PropValueIsSimple,       },
	{ PKEY_Document_DateSaved,               VT_FILETIME, S_OK, kXMP_NS_XMP, "ModifyDate",  kXMP_PropValueIsSimple,       },
};

/*!
** @brief   The fundamental data type by which property values are read and written.
** @remarks -
*/
struct CPropVariant : public PROPVARIANT
{
	/*!
	** @brief   Initializes a PROPVARIANT structure, and sets its type to VT_EMPTY.
	** @remarks -
	*/
	CPropVariant()
	{
		::PropVariantInit(this);
	}

	/*!
	** @brief   Frees all elements that can be freed in the PROPVARIANT structure.
	** @remarks -
	*/
	~CPropVariant()
	{
		ATLVERIFY(SUCCEEDED(::PropVariantClear(this)));
	}
};

/*!
** @brief   The COM object that implements the property handler.
** @remarks -
*/
class
DECLSPEC_UUID("E23F06B2-6619-46A5-87A9-F52B08E2A69A")
DECLSPEC_NOVTABLE
CPropertyHandler :
		public CComObjectRootEx<CComSingleThreadModel>,
		public CComCoClass<CPropertyHandler, &__uuidof(CPropertyHandler)>,
		public IPropertyStore,
		public IPropertyStoreCapabilities,
		public IInitializeWithStream,
		public IInitializeWithFile
{
private:
	CComPtr<IPropertyStoreCache> m_spCache;
	DWORD m_grfMode;
	SXMPFiles m_files;
	SXMPMeta m_meta;

	static CStringA UnicodeToUTF8(const CStringW& unicode)
	{
		CStringA utf8;
		if (unicode.IsEmpty())
			return utf8;
		int cbyUTF8 = ::WideCharToMultiByte(CP_UTF8, 0, unicode, unicode.GetLength(), NULL, 0, NULL, NULL);
		if (cbyUTF8 == 0)
			AtlThrowLastWin32();
		::WideCharToMultiByte(CP_UTF8, 0, unicode, unicode.GetLength(), utf8.GetBuffer(cbyUTF8), cbyUTF8, NULL, NULL);
		utf8.ReleaseBuffer(cbyUTF8);
		return utf8;
	}

	static CStringW UTF8ToUnicode(const CStringA& utf8)
	{
		CStringW unicode;
		if (utf8.IsEmpty())
			return unicode;
		int cchUnicode = ::MultiByteToWideChar(CP_UTF8, 0, utf8, utf8.GetLength(), NULL, 0);
		if (cchUnicode == 0)
			AtlThrowLastWin32();
		::MultiByteToWideChar(CP_UTF8, 0, utf8, utf8.GetLength(), unicode.GetBuffer(cchUnicode), cchUnicode);
		unicode.ReleaseBuffer(cchUnicode);
		return unicode;
	}

public:

	/*!
	** @brief   Constructs a newly allocated COM object.
	** @remarks -
	*/
	CPropertyHandler()
	{
		m_grfMode = STGM_READ;
	}

	DECLARE_REGISTRY_RESOURCEID(IDR_METADATA)

	BEGIN_COM_MAP(CPropertyHandler)
		COM_INTERFACE_ENTRY(IPropertyStore)
		COM_INTERFACE_ENTRY(IPropertyStoreCapabilities)
		COM_INTERFACE_ENTRY_NOINTERFACE(IInitializeWithStream)
		COM_INTERFACE_ENTRY(IInitializeWithFile)
	END_COM_MAP()

	DECLARE_PROTECT_FINAL_CONSTRUCT()

	/*!
	** @brief   Overrides to perform any initialization required for ths COM object.
	** @returns S_OK on success or one of the standard error HRESULT values.
	** @remarks -
	*/
	HRESULT FinalConstruct()
	{
		HRESULT hResult = S_OK;

		return hResult;
	}

	/*!
	** @brief   Overrides to perform any cleanup required for ths COM object.
	** @remarks -
	*/
	void FinalRelease()
	{
		;
	}

	/*!
	** @brief   Initializes a VT_FILETIME propvariant based on a specified string.
	** @param   [in] psz a null-terminated UTF-8 string in ISO 8601 format.
	** @param   [out] ppropvar an uninitialized PROPVARIANT structure that is initialized.
	** @returns S_OK if successful, or an HRESULT error code otherwise.
	** @remarks -
	*/
	static HRESULT InitPropFileTimeFromXMPString(XMP_StringPtr psz, PROPVARIANT* ppropvar)
	{
		::PropVariantInit(ppropvar); // to ensure initialization of the propvariant

		if (!psz)
			return E_INVALIDARG;
		if (!psz[0])
			return S_OK; // as VT_EMPTY

		HRESULT hResult = S_OK;

		XMP_DateTime time{};

		SXMPUtils::ConvertToDate(psz, &time);
		if (!time.hasTimeZone)
			SXMPUtils::SetTimeZone(&time);
		SXMPUtils::ConvertToUTCTime(&time);

		SYSTEMTIME st{};

		st.wYear   = WORD(time.year);
		st.wMonth  = WORD(time.month);
		st.wDay    = WORD(time.day);
		st.wHour   = WORD(time.hour);
		st.wMinute = WORD(time.minute);
		st.wSecond = WORD(time.second);

		FILETIME ft{};

		if (::SystemTimeToFileTime(&st, &ft))
		{
			hResult = ::InitPropVariantFromFileTime(&ft, ppropvar);
		}
		else
			hResult = AtlHresultFromLastError();

		return hResult;
	}

	/*!
	** @brief   Converts a VT_FILETIME propvariant to a ISO 8601 string.
	** @param   [in, out] ppropvar a PROPVARIANT structure to be coerced.
	** @returns S_OK if successful, or an HRESULT error code otherwise.
	** @remarks -
	*/
	static HRESULT CoerceFileTimeToISO8601String(PROPVARIANT *ppropvar)
	{
		if (!ppropvar)
			return E_INVALIDARG;
		if (ppropvar->vt != VT_FILETIME)
			return TYPE_E_TYPEMISMATCH;

		HRESULT hResult = S_OK;
		std::string strDateTime;

		FILETIME ft{};

		hResult = ::PropVariantToFileTime(*ppropvar, PSTF_UTC, &ft);

		if (SUCCEEDED(hResult))
		{
			SYSTEMTIME st{};

			if (::FileTimeToSystemTime(&ft, &st))
			{
				XMP_DateTime time{};

				time.year   = XMP_Int32(st.wYear);
				time.month  = XMP_Int32(st.wMonth);
				time.day    = XMP_Int32(st.wDay);
				time.hour   = XMP_Int32(st.wHour);
				time.minute = XMP_Int32(st.wMinute);
				time.second = XMP_Int32(st.wSecond);

				time.hasDate     = true;
				time.hasTime     = true;
				time.hasTimeZone = true;

				SXMPUtils::ConvertToLocalTime(&time);

				SXMPUtils::ConvertFromDate(time, &strDateTime);
			}
			else
				hResult = AtlHresultFromLastError();
		}
		if (SUCCEEDED(hResult))
			hResult = ::PropVariantClear(ppropvar);
		if (SUCCEEDED(hResult))
			hResult = ::InitPropVariantFromString(UTF8ToUnicode(strDateTime.c_str()), ppropvar);

		return hResult;
	}

	/*!
	** @brief   Reads the properties from the XMP metadata.
	** @param   [in] ppsCache the interface of an in-memory property store.
	** @param   [in] meta an XMP object from which to read the properties.
	** @returns No return value.
	** @remarks -
	*/
	static void LoadProperties(IPropertyStoreCache* ppsCache, const SXMPMeta& meta)
	{
		ATLASSERT(ppsCache);

		for (size_t iProp = 0, nProps = _countof(CProperty::values); iProp < nProps; ++iProp)
		{
			const CProperty& property = CProperty::values[iProp];

			HRESULT hResult = S_OK;

			XMP_StringPtr schemaNS = property.schemaNS;
			XMP_StringPtr propName = property.propName;
			std::string propValue;
			XMP_OptionBits options = kXMP_NoOptions;

			if (!meta.GetProperty(schemaNS, propName, &propValue, &options))
			{
				continue;
			}
			else if (XMP_PropIsSimple(options))
			{
				;
			}
			else if (XMP_PropIsArray(options) && XMP_ArrayIsAltText(options))
			{
				meta.GetLocalizedText(schemaNS, propName, NULL, "x-default", NULL, &propValue, NULL);
			}
			else if (XMP_PropIsArray(options) && XMP_ArrayIsAlternate(options))
			{
				;
			}
			else if (XMP_PropIsArray(options))
			{
				propValue.clear();

				for (XMP_Index iItem = 0, nItems = meta.CountArrayItems(schemaNS, propName); iItem < nItems; ++iItem)
				{
					std::string itemValue;

					meta.GetArrayItem(schemaNS, propName, 1 + iItem, &itemValue, NULL);

					if (itemValue.empty())
						continue;
					if (!propValue.empty())
						propValue += ';';

					propValue += itemValue;
				}
			}
			else
			{
				;
			}
			CPropVariant propvar;

			PSC_STATE state = property.writable == S_FALSE ? PSC_READONLY : PSC_NORMAL;

			switch (property.vartype)
			{
			case VT_FILETIME:
				hResult = InitPropFileTimeFromXMPString(propValue.c_str(), &propvar);
				break;
			default:
				hResult = ::InitPropVariantFromString(UTF8ToUnicode(propValue.c_str()), &propvar);
				break;
			}
			ATLENSURE_SUCCEEDED(hResult);
			hResult = ::PSCoerceToCanonicalValue(property.propkey, &propvar);
			ATLENSURE_SUCCEEDED(hResult);
			hResult = ppsCache->SetValueAndState(property.propkey, &propvar, state);
			ATLENSURE_SUCCEEDED(hResult);
		}
	}

	/*!
	** @brief   Writes the modified properties back to the XMP metadata.
	** @param   [in] ppsCache the interface of an in-memory property store.
	** @param   [out] meta an XMP object to which to write the properties.
	** @returns true if any property has changed, or false otherwise.
	** @remarks -
	*/
	static bool SaveProperties(IPropertyStoreCache* ppsCache, SXMPMeta& meta)
	{
		ATLASSERT(ppsCache);

		bool bDirty = false;

		for (size_t iProp = 0, nProps = _countof(CProperty::values); iProp < nProps; ++iProp)
		{
			const CProperty& property = CProperty::values[iProp];

			HRESULT hResult = S_OK;

			XMP_StringPtr schemaNS = property.schemaNS;
			XMP_StringPtr propName = property.propName;
			XMP_OptionBits options = property.options;

			CPropVariant propvar;
			PSC_STATE state;
			switch (hResult = ppsCache->GetValueAndState(property.propkey, &propvar, &state))
			{
			case TYPE_E_ELEMENTNOTFOUND:
				continue;
			default:
				break;
			}
			ATLENSURE_SUCCEEDED(hResult);
			hResult = ::PSCoerceToCanonicalValue(property.propkey, &propvar);
			ATLENSURE_SUCCEEDED(hResult);
			if (state != PSC_DIRTY)
				continue;

			if (propvar.vt == VT_EMPTY)
			{
				meta.DeleteProperty(schemaNS, propName);
			}
			else if (XMP_PropIsSimple(options))
			{
				CStringW propValue;

				for (ULONG iElem = 0, nElems = ::PropVariantGetElementCount(propvar); iElem < nElems; ++iElem)
				{
					CPropVariant propvarItem;
					hResult = ::PropVariantGetElem(propvar, iElem, &propvarItem);
					ATLENSURE_SUCCEEDED(hResult);
					switch (propvarItem.vt)
					{
					case VT_FILETIME:
						hResult = CoerceFileTimeToISO8601String(&propvarItem);
						ATLENSURE_SUCCEEDED(hResult);
						break;
					default:
						break;
					}
					ATLASSERT(propvarItem.vt == VT_LPWSTR);
					LPWSTR itemValue = propvarItem.pwszVal;

					if (!propValue.IsEmpty())
						propValue += L';';

					propValue += itemValue;
				}
				meta.SetProperty(schemaNS, propName, UnicodeToUTF8(propValue));
			}
			else if (XMP_PropIsArray(options) && XMP_ArrayIsAltText(options))
			{
				meta.SetProperty(schemaNS, propName, NULL, options);

				ATLASSERT(propvar.vt == VT_LPWSTR);
				LPWSTR propValue = propvar.pwszVal;

				meta.SetLocalizedText(schemaNS, propName, NULL, "x-default", UnicodeToUTF8(propValue));
			}
			else if (XMP_PropIsArray(options) && XMP_ArrayIsAlternate(options))
			{
				;
			}
			else if (XMP_PropIsArray(options))
			{
				meta.SetProperty(schemaNS, propName, NULL, options);

				for (ULONG iElem = 0, nElems = ::PropVariantGetElementCount(propvar); iElem < nElems; ++iElem)
				{
					CPropVariant propvarItem;
					hResult = ::PropVariantGetElem(propvar, iElem, &propvarItem);
					ATLENSURE_SUCCEEDED(hResult);
					ATLASSERT(propvarItem.vt == VT_LPWSTR);
					LPWSTR itemValue = propvarItem.pwszVal;

					meta.AppendArrayItem(schemaNS, propName, options, UnicodeToUTF8(itemValue));
				}
			}
			else
			{
				;
			}
			#if 1 // Automatically update xmp:MetadataDate

			if (IsEqualPropertyKey(property.propkey, PKEY_Document_DateSaved))
			{
				if (meta.DoesPropertyExist(kXMP_NS_XMP, "MetadataDate"))
				{
					std::string propValue;

					if (meta.GetProperty(kXMP_NS_XMP, "ModifyDate", &propValue, NULL))
						meta.SetProperty(kXMP_NS_XMP, "MetadataDate", propValue);
					#if 0 // delete xmp:MetadataDate if no xmp:ModifyDate exists.
					else
						meta.DeleteProperty(kXMP_NS_XMP, "MetadataDate");
					#endif
				}
			}
			#endif
			bDirty = true;
		}
		return bDirty;
	}

	/// @name IPropertyStore interface
	/// Exposes methods used to enumerate and manipulate property values.
	/// @{

	/*!
	** @brief   Retrieves the number of properties that are attached to the file.
	** @param   [out] pcProps a pointer to a value that indicates the property count.
	** @returns S_OK on success or one of the standard error HRESULT values.
	** @remarks -
	*/
	HRESULT STDMETHODCALLTYPE GetCount(DWORD* pcProps)
	{
		if (!m_spCache)
			return E_UNEXPECTED;

		ATLASSERT(m_spCache);
		HRESULT hResult = m_spCache->GetCount(pcProps);
		return hResult;
	}

	/*!
	** @brief   Gets a property key from the property array of an item.
	** @param   [in] iProp the zero-based index of the property key in the array.
	** @param   [out] pkey the address of a variable that receives the property key.
	** @returns S_OK on success or one of the standard error HRESULT values.
	** @remarks -
	*/
	HRESULT STDMETHODCALLTYPE GetAt(DWORD iProp, PROPERTYKEY* pkey)
	{
		if (!m_spCache)
			return E_UNEXPECTED;

		ATLASSERT(m_spCache);
		HRESULT hResult = m_spCache->GetAt(iProp, pkey);
		return hResult;
	}

	/*!
	** @brief   Retrieves the data for a specific property.
	** @param   [in] key a property key that identifies the requested property.
	** @param   [out] ppropvar a pointer to an uninitialized PROPVARIANT structure.
	** @returns S_OK or INPLACE_S_TRUNCATED on success, or an error value otherwise.
	** @remarks -
	*/
	HRESULT STDMETHODCALLTYPE GetValue(REFPROPERTYKEY key, PROPVARIANT* ppropvar)
	{
		if (!m_spCache)
			return E_UNEXPECTED;

		ATLASSERT(m_spCache);
		HRESULT hResult = m_spCache->GetValue(key, ppropvar);
		return hResult;
	}

	/*!
	** @brief   Sets a property value or replaces or removes an existing value.
	** @param   [in] key a property key that identifies the requested property.
	** @param   [in] propvar a PROPVARIANT structure that contains the new value.
	** @returns S_OK or INPLACE_S_TRUNCATED on success, or an error value otherwise.
	** @remarks -
	*/
	HRESULT STDMETHODCALLTYPE SetValue(REFPROPERTYKEY key, REFPROPVARIANT propvar)
	{
		if (!m_spCache)
			return E_UNEXPECTED;
		if (!m_files.GetFileInfo())
			return E_UNEXPECTED;
		if (!(m_grfMode & STGM_READWRITE))
			return STG_E_ACCESSDENIED;
		if (IsPropertyWritable(key) == S_FALSE)
			return E_INVALIDARG;

		ATLASSERT(m_spCache);
		HRESULT hResult = m_spCache->SetValueAndState(key, &propvar, PSC_DIRTY);
		return hResult;
	}

	/*!
	** @brief   After a change has been made, this method saves the changes.
	** @returns S_OK on success or one of the standard error HRESULT values.
	** @remarks -
	*/
	HRESULT STDMETHODCALLTYPE Commit()
	{
		if (!m_spCache)
			return E_UNEXPECTED;
		if (!m_files.GetFileInfo())
			return E_UNEXPECTED;
		if (!(m_grfMode & STGM_READWRITE))
			return STG_E_ACCESSDENIED;

		HRESULT hResult = S_OK;
		_ATLTRY
		{
			try
			{
				bool bDirty = SaveProperties(m_spCache, m_meta);

				if (bDirty)
				{
					#if 0 // Automatically update xmpMM:InstanceID

					if (m_meta.DoesPropertyExist(kXMP_NS_XMP_MM, "InstanceID"))
					{
						GUID guid;
						hResult = ::CoCreateGuid(&guid);
						ATLENSURE_SUCCEEDED(hResult);
						CStringW strGUID;
						::StringFromGUID2(guid, strGUID.GetBuffer(38), 38 + 1);
						strGUID.ReleaseBuffer();
						strGUID.TrimLeft(L'{').TrimRight(L'}');
						CStringW propValue = L"uuid:" + strGUID;

						m_meta.SetProperty(kXMP_NS_XMP_MM, "InstanceID", UnicodeToUTF8(propValue));
					}
					#endif

					if (!m_files.CanPutXMP(m_meta))
						AtlThrow(STG_E_CANTSAVE);

					m_files.PutXMP(m_meta);
				}
				m_files.CloseFile();
			}
			catch (const std::bad_alloc& ex)
			{
				ATLTRACE2(atlTraceCOM, 0, TEXT("bad_alloc: %hs\n"), ex.what());

				AtlThrow(STG_E_INSUFFICIENTMEMORY);
			}
			catch (const XMP_Error& error)
			{
				ATLTRACE2(atlTraceCOM, 0, TEXT("XMP_Error: %hs\n"), error.GetErrMsg());

				switch (error.GetID())
				{
				case kXMPErr_NoMemory:
					AtlThrow(STG_E_INSUFFICIENTMEMORY);
				case kXMPErr_FilePermission:
					AtlThrow(STG_E_ACCESSDENIED);
				case kXMPErr_DiskSpace:
					AtlThrow(STG_E_MEDIUMFULL);
				default:
					AtlThrow(STG_E_UNKNOWN);
				}
			}
		}
		_ATLCATCH(ex)
		{
			hResult = ex;
		}
		_ATLCATCHALL()
		{
			hResult = E_FAIL;
		}
		return hResult;
	}

	/// @}

	/// @name IPropertyStoreCapabilities interface
	/// Exposes a method that determines whether a property can be edited in the UI by the user.
	/// @{

	/*!
	** @brief   Queries whether a specific property can be edited in the UI by the user.
	** @param   [in] key a property key that represents the property being queried.
	** @returns S_OK if the property can be edited, or S_FALSE otherwise.
	** @remarks -
	*/
	HRESULT STDMETHODCALLTYPE IsPropertyWritable(REFPROPERTYKEY key)
	{
		for (size_t iProp = 0, nProps = _countof(CProperty::values); iProp < nProps; ++iProp)
		{
			const CProperty& property = CProperty::values[iProp];
			if (IsEqualPropertyKey(key, property.propkey))
				return property.writable;
		}
		return S_FALSE;
	}

	/// @}

	/// @name IInitializeWithStream interface
	/// Exposes a method that initializes a handler with a stream.
	/// @{

	/*!
	** @brief   Initializes a handler with a stream.
	** @param   [in] pStream an IStream interface that represents the stream source.
	** @param   [in] grfMode the STGM flags that indicates the access mode.
	** @returns S_OK on success or one of the standard error HRESULT values.
	** @remarks Not implemented.
	*/
	HRESULT STDMETHODCALLTYPE Initialize(IStream* pStream, DWORD grfMode)
	{
		UNREFERENCED_PARAMETER(pStream);
		UNREFERENCED_PARAMETER(grfMode);

		return E_NOTIMPL;
	}

	/// @}

	/// @name IInitializeWithFile interface
	/// Exposes a method that initializes a handler with a file path.
	/// @{

	/*!
	** @brief   Initializes a handler with a file path.
	** @param   [in] pszFilePath the file path as a null-terminated Unicode string.
	** @param   [in] grfMode the STGM flags that indicates the access mode.
	** @returns S_OK on success or one of the standard error HRESULT values.
	** @remarks -
	*/
	HRESULT STDMETHODCALLTYPE Initialize(LPCWSTR pszFilePath, DWORD grfMode)
	{
		if (m_spCache)
			return HRESULT_FROM_WIN32(ERROR_ALREADY_INITIALIZED);
		if (!pszFilePath
		 || !pszFilePath[0])
			return E_INVALIDARG;

		ATLTRACE2(atlTraceCOM, 0, TEXT("IInitializeWithFile::Initialize(\"%ls\", 0x%08X);\n"), pszFilePath, grfMode);

		HRESULT hResult = S_OK;
		_ATLTRY
		{
			CComPtr<IPropertyStoreCache> spCache;
			hResult = ::PSCreateMemoryPropertyStore(IID_PPV_ARGS(&spCache));
			ATLENSURE_SUCCEEDED(hResult);
			ATLASSERT(spCache);

			try
			{
				XMP_OptionBits openFlags;
				if (grfMode & STGM_READWRITE)
					openFlags = kXMPFiles_OpenForUpdate;
				else
					openFlags = kXMPFiles_OpenForRead;

				if (!m_files.OpenFile(UnicodeToUTF8(pszFilePath), kXMP_PDFFile, openFlags))
					AtlThrow(E_FAIL);
				if (!m_files.GetXMP(&m_meta))
					AtlThrow(E_FAIL);

				LoadProperties(spCache, m_meta);
			}
			catch (const std::bad_alloc& ex)
			{
				ATLTRACE2(atlTraceCOM, 0, TEXT("bad_alloc: %hs\n"), ex.what());

				AtlThrow(STG_E_INSUFFICIENTMEMORY);
			}
			catch (const XMP_Error& error)
			{
				ATLTRACE2(atlTraceCOM, 0, TEXT("XMP_Error: %hs\n"), error.GetErrMsg());

				switch (error.GetID())
				{
				case kXMPErr_NoMemory:
					AtlThrow(STG_E_INSUFFICIENTMEMORY);
				case kXMPErr_FilePermission:
					AtlThrow(STG_E_ACCESSDENIED);
				case kXMPErr_DiskSpace:
					AtlThrow(STG_E_MEDIUMFULL);
				default:
					AtlThrow(STG_E_UNKNOWN);
				}
			}
			m_spCache = spCache;
			m_grfMode = grfMode;
		}
		_ATLCATCH(ex)
		{
			hResult = ex;

			m_files.CloseFile(), m_meta.Erase();
		}
		_ATLCATCHALL()
		{
			hResult = E_FAIL;

			m_files.CloseFile(), m_meta.Erase();
		}
		return hResult;
	}

	/// @}
};

OBJECT_ENTRY_AUTO(__uuidof(CPropertyHandler), CPropertyHandler)
