/*!@file
** @brief   The module-definition of the dynamic-link library.
** @remarks This file is the single compilation unit of the program.
** @author  UETA, Shin-ichi <ueta@usdesign.jp>
**
** @copyright (C) 2019 Ueta System Design Studio\n
** This file is licensed under the terms of the MIT License.\n
** See <https://opensource.org/licenses/MIT>.
*/
#include "stdafx.h"
#include "resource.h"
#include "module.h"
#include "metadata.h"

/*!
** @brief   The global instance of the ATL module.
** @remarks -
*/
CAtlDllModule _AtlModule;

/*!
** @brief   Implements the entry point into the dynamic-link library.
** @param   [in] hInstance a handle to the DLL module.
** @param   [in] dwReason the reason code that indicates why the entry point is being called.
** @param   [in] lpReserved see the reference documentation for the details.
** @returns TRUE if successful, or FALSE otherwise.
** @remarks -
*/
extern "C" BOOL WINAPI DllMain(HINSTANCE hInstance, DWORD dwReason, LPVOID lpReserved)
{
	UNREFERENCED_PARAMETER(hInstance);

	BOOL bSucceeded = _AtlModule.DllMain(dwReason, lpReserved);
	if (bSucceeded)
	{
		switch (dwReason)
		{
		case DLL_PROCESS_ATTACH:
			if (!SXMPMeta::Initialize()
			 || !SXMPFiles::Initialize())
			{
				bSucceeded = FALSE;
			}
			break;
		case DLL_PROCESS_DETACH:
			SXMPFiles::Terminate();
			SXMPMeta::Terminate();
			break;
		default:
			break;
		}
	}
	return bSucceeded;
}

/*!
** @brief   Determines whether the DLL that implements this function is in use.
** @returns S_OK if the DLL can be unloaded, or S_FALSE if it cannot.
** @remarks -
*/
STDAPI DllCanUnloadNow()
{
	HRESULT hResult = _AtlModule.DllCanUnloadNow();
	return hResult;
}

/*!
** @brief   Retrieves the class object from a DLL object handler or object application.
** @param   [in] rclsid the CLSID that will associate the correct data and code.
** @param   [in] riid the identifier of the interface to be used to communicate with the object.
** @param   [out] ppv the address of a pointer variable that receives the interface pointer.
** @returns S_OK on success or one of the standard error HRESULT values.
** @remarks -
*/
STDAPI DllGetClassObject(REFCLSID rclsid, REFIID riid, LPVOID* ppv)
{
	HRESULT hResult = _AtlModule.DllGetClassObject(rclsid, riid, ppv);
	return hResult;
}

/*!
** @brief   Instructs an in-process server to create its registry entries.
** @returns S_OK on success or one of the standard error HRESULT values.
** @remarks -
*/
STDAPI DllRegisterServer()
{
	HRESULT hResult = _AtlModule.DllRegisterServer(FALSE);
	return hResult;
}

/*!
** @brief   Instructs an in-process server to remove its registry entries.
** @returns S_OK on success or one of the standard error HRESULT values.
** @remarks -
*/
STDAPI DllUnregisterServer()
{
	HRESULT hResult = _AtlModule.DllUnregisterServer(FALSE);
	return hResult;
}

/*!
** @brief   Handles installation and setup for an in-process server.
** @param   [in] bInstall TRUE if being installed, or FALSE if being uninstalled.
** @param   [in] pszCmdLine a string passed in by regsvr32 with the /i option.
** @returns S_OK on success or one of the standard error HRESULT values.
** @remarks -
*/
STDAPI DllInstall(BOOL bInstall, LPCWSTR pszCmdLine)
{
	static const wchar_t szUserSwitch[] = L"user";

	HRESULT hResult = E_FAIL;

	if (pszCmdLine)
	{
		if (_wcsnicmp(pszCmdLine, szUserSwitch, _countof(szUserSwitch)) == 0)
		{
			AtlSetPerUserRegistration(true);
		}
	}
	if (bInstall)
	{
		hResult = DllRegisterServer();

		if (FAILED(hResult))
		{
			DllUnregisterServer();
		}
	}
	else
	{
		hResult = DllUnregisterServer();
	}
	return hResult;
}
