/*!@file
** @brief   The test driver for the in-process COM server.
** @remarks The test driver can be called by the rundll32 command.
** @author  UETA, Shin-ichi <ueta@usdesign.jp>
**
** @copyright (C) 2019 Ueta System Design Studio\n
** This file is licensed under the terms of the MIT License.\n
** See <https://opensource.org/licenses/MIT>.
*/
#include "stdafx.h"
#include "resource.h"
#include "module.h"
#include "metadata.h"

namespace
{
	/*!
	** @brief   Creates an empty file.
	** @param   [in] lpFileName the name of the file to be created.
	** @returns TRUE if successful, or FALSE otherwise.
	** @remarks -
	*/
	BOOL CreateFile(LPCTSTR lpFileName)
	{
		HANDLE hFile = ::CreateFile(
				lpFileName,
				GENERIC_READ | GENERIC_WRITE,
				0,
				NULL,
				CREATE_ALWAYS,
				FILE_ATTRIBUTE_NORMAL,
				NULL);

		if (hFile != INVALID_HANDLE_VALUE)
		{
			::CloseHandle(hFile);

			return TRUE;
		}
		return FALSE;
	}
};

/*!
** @brief   Implements the test driver for the in-process COM server.
** @param   [in] hWnd a window handle that should be used as the owner window.
** @param   [in] hInstance a handle to the DLL module.
** @param   [in] lpszCmdLine a string passed in by the rundll32 command.
** @param   [in] nCmdShow the flags that describes how the DLL's windows should be displayed.
** @returns No return value.
** @remarks -
*/
extern "C" __declspec(dllexport) void CALLBACK test(HWND hWnd, HINSTANCE hInstance, LPSTR lpszCmdLine, int nCmdShow)
{
	UNREFERENCED_PARAMETER(hWnd);
	UNREFERENCED_PARAMETER(hInstance);
	UNREFERENCED_PARAMETER(lpszCmdLine);
	UNREFERENCED_PARAMETER(nCmdShow);

	_CrtSetDbgFlag(_CRTDBG_ALLOC_MEM_DF | _CRTDBG_LEAK_CHECK_DF);

	HRESULT hResult = S_OK;
	hResult = ::CoInitialize(NULL);
	ATLASSERT(SUCCEEDED(hResult));
	PROPERTYKEY propkey = PKEY_Null;
	PROPVARIANT propvar;
	::PropVariantInit(&propvar);

	SYSTEMTIME stJustNow{};
	::GetSystemTime(&stJustNow), stJustNow.wMilliseconds = 0;
	FILETIME ftJustNow{};
	::SystemTimeToFileTime(&stJustNow, &ftJustNow);

	CComPtr<IClassFactory> spClassFactory;
	hResult = ::DllGetClassObject(__uuidof(CPropertyHandler), IID_PPV_ARGS(&spClassFactory));
	ATLASSERT(SUCCEEDED(hResult));

	#if 1 // Setup
	{
		ATLTRACE(TEXT("Setup\n"));

		ATLVERIFY(::CopyFile(TEXT("src/sample.pdf"), TEXT("target/normal.pdf"), FALSE));
		ATLVERIFY(::SetFileAttributes(TEXT("target/normal.pdf"), FILE_ATTRIBUTE_NORMAL));
		ATLVERIFY(::CopyFile(TEXT("src/sample.pdf"), TEXT("target/read-only.pdf"), FALSE));
		ATLVERIFY(::SetFileAttributes(TEXT("target/read-only.pdf"), FILE_ATTRIBUTE_READONLY));
		ATLVERIFY(CreateFile(TEXT("target/empty.txt")));
		ATLVERIFY(::SetFileAttributes(TEXT("target/empty.txt"), FILE_ATTRIBUTE_NORMAL));
	}
	#endif

	#if 1 // Test 1: IPropertyStoreCapabilities
	{
		ATLTRACE(TEXT("1:Start\n"));

		CComPtr<IPropertyStore> spPropertyStore;
		hResult = spClassFactory->CreateInstance(NULL, IID_PPV_ARGS(&spPropertyStore));
		ATLASSERT(SUCCEEDED(hResult));
		CComQIPtr<IPropertyStoreCapabilities> spPropertyStoreCapabilities = spPropertyStore;
		ATLASSERT(spPropertyStoreCapabilities);

		hResult = spPropertyStoreCapabilities->IsPropertyWritable(PKEY_Title);
		ATLASSERT(hResult == S_OK);
		hResult = spPropertyStoreCapabilities->IsPropertyWritable(PKEY_Subject);
		ATLASSERT(hResult == S_OK);
		hResult = spPropertyStoreCapabilities->IsPropertyWritable(PKEY_Author);
		ATLASSERT(hResult == S_OK);
		hResult = spPropertyStoreCapabilities->IsPropertyWritable(PKEY_Keywords);
		ATLASSERT(hResult == S_OK);
		hResult = spPropertyStoreCapabilities->IsPropertyWritable(PKEY_Comment);
		ATLASSERT(hResult == S_FALSE);
		hResult = spPropertyStoreCapabilities->IsPropertyWritable(PKEY_Document_TotalEditingTime);
		ATLASSERT(hResult == S_FALSE);
		hResult = spPropertyStoreCapabilities->IsPropertyWritable(PKEY_Document_DatePrinted);
		ATLASSERT(hResult == S_FALSE);
		hResult = spPropertyStoreCapabilities->IsPropertyWritable(PKEY_Document_DateCreated);
		ATLASSERT(hResult == S_OK);
		hResult = spPropertyStoreCapabilities->IsPropertyWritable(PKEY_Document_DateSaved);
		ATLASSERT(hResult == S_OK);
		hResult = spPropertyStoreCapabilities->IsPropertyWritable(PKEY_ApplicationName);
		ATLASSERT(hResult == S_OK);

		spPropertyStoreCapabilities.Release();

		spPropertyStore.Release();

		::Sleep(300);

		ATLTRACE(TEXT("1:Done\n"));
	}
	#endif

	#if 1 // Test 2: IInitializeWithStream
	{
		ATLTRACE(TEXT("2:Start\n"));

		CComPtr<IPropertyStore> spPropertyStore;
		hResult = spClassFactory->CreateInstance(NULL, IID_PPV_ARGS(&spPropertyStore));
		ATLASSERT(SUCCEEDED(hResult));
		CComPtr<IInitializeWithStream> spInitializeWithStream;
		hResult = spPropertyStore->QueryInterface(&spInitializeWithStream);
		// Not implemented, and so prevent the interface from being used.
		ATLASSERT(hResult == E_NOINTERFACE);

		spPropertyStore.Release();

		::Sleep(300);

		ATLTRACE(TEXT("2:Done\n"));
	}
	#endif

	#if 1 // Test 3: IInitializeWithFile
	{
		ATLTRACE(TEXT("3:Start\n"));

		CComPtr<IPropertyStore> spPropertyStore;
		hResult = spClassFactory->CreateInstance(NULL, IID_PPV_ARGS(&spPropertyStore));
		ATLASSERT(SUCCEEDED(hResult));
		CComQIPtr<IInitializeWithFile> spInitializeWithFile = spPropertyStore;
		ATLASSERT(spInitializeWithFile);

		hResult = spInitializeWithFile->Initialize(NULL, STGM_READWRITE);
		ATLASSERT(hResult == E_INVALIDARG);
		hResult = spInitializeWithFile->Initialize(L"", STGM_READWRITE);
		ATLASSERT(hResult == E_INVALIDARG);
		hResult = spInitializeWithFile->Initialize(L"target/phantom.pdf", STGM_READWRITE);
		ATLASSERT(hResult == E_FAIL);
		hResult = spInitializeWithFile->Initialize(L"target/empty.txt", STGM_READWRITE);
		ATLASSERT(hResult == E_FAIL);
		hResult = spInitializeWithFile->Initialize(L"target/normal.pdf", STGM_READWRITE);
		ATLASSERT(SUCCEEDED(hResult));
		hResult = spInitializeWithFile->Initialize(L"target/normal.pdf", STGM_READWRITE);
		ATLASSERT(hResult == HRESULT_FROM_WIN32(ERROR_ALREADY_INITIALIZED));

		spInitializeWithFile.Release();

		spPropertyStore.Release();

		::Sleep(300);

		ATLTRACE(TEXT("3:Done\n"));
	}
	#endif

	#if 1 // Test 4: IPropertyStore::GetCount and IPropertyStore::GetAt
	{
		ATLTRACE(TEXT("4:Start\n"));

		DWORD cProps = 0;

		CComPtr<IPropertyStore> spPropertyStore;
		hResult = spClassFactory->CreateInstance(NULL, IID_PPV_ARGS(&spPropertyStore));
		ATLASSERT(SUCCEEDED(hResult));
		hResult = spPropertyStore->GetCount(&cProps);
		ATLASSERT(hResult == E_UNEXPECTED);
		hResult = spPropertyStore->GetAt(0, &propkey);
		ATLASSERT(hResult == E_UNEXPECTED);

		CComQIPtr<IInitializeWithFile> spInitializeWithFile = spPropertyStore;
		ATLASSERT(spInitializeWithFile);
		hResult = spInitializeWithFile->Initialize(L"target/normal.pdf", STGM_READ);
		ATLASSERT(SUCCEEDED(hResult));
		spInitializeWithFile.Release();

		hResult = spPropertyStore->GetCount(NULL);
		ATLASSERT(hResult == E_POINTER);
		hResult = spPropertyStore->GetCount(&cProps);
		ATLASSERT(SUCCEEDED(hResult));

		ATLASSERT(cProps == 7);

		for (DWORD iProp = 0; iProp < cProps; ++iProp)
		{
			propkey = PKEY_Null;

			hResult = spPropertyStore->GetAt(iProp, &propkey);
			ATLASSERT(SUCCEEDED(hResult));
			ATLASSERT(!IsEqualPropertyKey(propkey, PKEY_Null));
		}
		propkey = PKEY_Null;

		hResult = spPropertyStore->GetAt(cProps, &propkey);
		ATLASSERT(hResult == E_INVALIDARG);
		hResult = spPropertyStore->GetAt(0, NULL);
		ATLASSERT(hResult == E_POINTER);
		spPropertyStore.Release();

		::Sleep(300);

		ATLTRACE(TEXT("4:Done\n"));
	}
	#endif

	#if 1 // Test 5: IPropertyStore::GetValue
	{
		ATLTRACE(TEXT("5:Start\n"));

		CComPtr<IPropertyStore> spPropertyStore;
		hResult = spClassFactory->CreateInstance(NULL, IID_PPV_ARGS(&spPropertyStore));
		ATLASSERT(SUCCEEDED(hResult));
		hResult = spPropertyStore->GetValue(PKEY_Null, &propvar);
		ATLASSERT(hResult == E_UNEXPECTED);

		CComQIPtr<IInitializeWithFile> spInitializeWithFile = spPropertyStore;
		ATLASSERT(spInitializeWithFile);
		hResult = spInitializeWithFile->Initialize(L"target/normal.pdf", STGM_READ);
		ATLASSERT(SUCCEEDED(hResult));
		spInitializeWithFile.Release();

		hResult = spPropertyStore->GetValue(PKEY_Title, &propvar);
		ATLASSERT(SUCCEEDED(hResult));
		ATLASSERT(propvar.vt == VT_LPWSTR);
		ATLASSERT(wcscmp(propvar.pwszVal, L"title") == 0);
		::PropVariantClear(&propvar);

		hResult = spPropertyStore->GetValue(PKEY_Subject, &propvar);
		ATLASSERT(SUCCEEDED(hResult));
		ATLASSERT(propvar.vt == VT_LPWSTR);
		ATLASSERT(wcscmp(propvar.pwszVal, L"subject") == 0);
		::PropVariantClear(&propvar);

		hResult = spPropertyStore->GetValue(PKEY_Author, &propvar);
		ATLASSERT(SUCCEEDED(hResult));
		ATLASSERT(propvar.vt == (VT_VECTOR | VT_LPWSTR));
		ATLASSERT(propvar.calpwstr.cElems == 2);
		ATLASSERT(wcscmp(propvar.calpwstr.pElems[0], L"author1") == 0);
		ATLASSERT(wcscmp(propvar.calpwstr.pElems[1], L"author2") == 0);
		::PropVariantClear(&propvar);

		hResult = spPropertyStore->GetValue(PKEY_Keywords, &propvar);
		ATLASSERT(SUCCEEDED(hResult));
		ATLASSERT(propvar.vt == (VT_VECTOR | VT_LPWSTR));
		ATLASSERT(propvar.calpwstr.cElems == 2);
		ATLASSERT(wcscmp(propvar.calpwstr.pElems[0], L"keyword1") == 0);
		ATLASSERT(wcscmp(propvar.calpwstr.pElems[1], L"keyword2") == 0);
		::PropVariantClear(&propvar);

		hResult = spPropertyStore->GetValue(PKEY_Comment, &propvar);
		ATLASSERT(SUCCEEDED(hResult));
		ATLASSERT(propvar.vt == VT_EMPTY);
		::PropVariantClear(&propvar);

		hResult = spPropertyStore->GetValue(PKEY_Document_TotalEditingTime, &propvar);
		ATLASSERT(SUCCEEDED(hResult));
		ATLASSERT(propvar.vt == VT_EMPTY);
		::PropVariantClear(&propvar);

		hResult = spPropertyStore->GetValue(PKEY_Document_DatePrinted, &propvar);
		ATLASSERT(SUCCEEDED(hResult));
		ATLASSERT(propvar.vt == VT_EMPTY);
		::PropVariantClear(&propvar);

		hResult = spPropertyStore->GetValue(PKEY_Document_DateCreated, &propvar);
		ATLASSERT(SUCCEEDED(hResult));
		ATLASSERT(propvar.vt == VT_FILETIME);
		::PropVariantClear(&propvar);

		hResult = spPropertyStore->GetValue(PKEY_Document_DateSaved, &propvar);
		ATLASSERT(SUCCEEDED(hResult));
		ATLASSERT(propvar.vt == VT_FILETIME);
		::PropVariantClear(&propvar);

		hResult = spPropertyStore->GetValue(PKEY_ApplicationName, &propvar);
		ATLASSERT(SUCCEEDED(hResult));
		ATLASSERT(propvar.vt == VT_LPWSTR);
		ATLASSERT(wcscmp(propvar.pwszVal, L"creator") == 0);
		::PropVariantClear(&propvar);

		hResult = spPropertyStore->GetValue(PKEY_Null, NULL);
		ATLASSERT(hResult == E_POINTER);
		spPropertyStore.Release();

		::Sleep(300);

		ATLTRACE(TEXT("5:Done\n"));
	}
	#endif

	#if 1 // Test 6: IPropertyStore::SetValue and IPropertyStore::Commit
	{
		ATLTRACE(TEXT("6:Start\n"));

		CComPtr<IPropertyStore> spPropertyStore;
		hResult = spClassFactory->CreateInstance(NULL, IID_PPV_ARGS(&spPropertyStore));
		ATLASSERT(SUCCEEDED(hResult));
		hResult = spPropertyStore->SetValue(PKEY_Null, propvar);
		ATLASSERT(hResult == E_UNEXPECTED);
		hResult = spPropertyStore->Commit();
		ATLASSERT(hResult == E_UNEXPECTED);

		CComQIPtr<IInitializeWithFile> spInitializeWithFile = spPropertyStore;
		ATLASSERT(spInitializeWithFile);
		hResult = spInitializeWithFile->Initialize(L"target/normal.pdf", STGM_READWRITE);
		ATLASSERT(SUCCEEDED(hResult));
		spInitializeWithFile.Release();

		hResult = ::InitPropVariantFromString(L"TITLE", &propvar);
		hResult = ::PSCoerceToCanonicalValue(PKEY_Title, &propvar);
		hResult = spPropertyStore->SetValue(PKEY_Title, propvar);
		ATLASSERT(SUCCEEDED(hResult));
		::PropVariantClear(&propvar);

		hResult = ::InitPropVariantFromString(L"SUBJECT", &propvar);
		hResult = ::PSCoerceToCanonicalValue(PKEY_Subject, &propvar);
		hResult = spPropertyStore->SetValue(PKEY_Subject, propvar);
		ATLASSERT(SUCCEEDED(hResult));
		::PropVariantClear(&propvar);

		hResult = ::InitPropVariantFromStringAsVector(L"AUTHOR1;AUTHOR2", &propvar);
		hResult = ::PSCoerceToCanonicalValue(PKEY_Author, &propvar);
		hResult = spPropertyStore->SetValue(PKEY_Author, propvar);
		ATLASSERT(SUCCEEDED(hResult));
		::PropVariantClear(&propvar);

		hResult = ::InitPropVariantFromStringAsVector(L"KEYWORD1;KEYWORD2", &propvar);
		hResult = ::PSCoerceToCanonicalValue(PKEY_Keywords, &propvar);
		hResult = spPropertyStore->SetValue(PKEY_Keywords, propvar);
		ATLASSERT(SUCCEEDED(hResult));
		::PropVariantClear(&propvar);

		hResult = ::InitPropVariantFromString(L"COMMENT", &propvar);
		hResult = ::PSCoerceToCanonicalValue(PKEY_Comment, &propvar);
		hResult = spPropertyStore->SetValue(PKEY_Comment, propvar);
		ATLASSERT(hResult == E_INVALIDARG);
		::PropVariantClear(&propvar);

		hResult = ::InitPropVariantFromUInt64(0, &propvar);
		hResult = ::PSCoerceToCanonicalValue(PKEY_Document_TotalEditingTime, &propvar);
		hResult = spPropertyStore->SetValue(PKEY_Document_TotalEditingTime, propvar);
		ATLASSERT(hResult == E_INVALIDARG);
		::PropVariantClear(&propvar);

		hResult = ::InitPropVariantFromFileTime(&ftJustNow, &propvar);
		hResult = ::PSCoerceToCanonicalValue(PKEY_Document_DatePrinted, &propvar);
		hResult = spPropertyStore->SetValue(PKEY_Document_DatePrinted, propvar);
		ATLASSERT(hResult == E_INVALIDARG);
		::PropVariantClear(&propvar);

		hResult = ::InitPropVariantFromFileTime(&ftJustNow, &propvar);
		hResult = ::PSCoerceToCanonicalValue(PKEY_Document_DateCreated, &propvar);
		hResult = spPropertyStore->SetValue(PKEY_Document_DateCreated, propvar);
		ATLASSERT(SUCCEEDED(hResult));
		::PropVariantClear(&propvar);

		hResult = ::InitPropVariantFromFileTime(&ftJustNow, &propvar);
		hResult = ::PSCoerceToCanonicalValue(PKEY_Document_DateSaved, &propvar);
		hResult = spPropertyStore->SetValue(PKEY_Document_DateSaved, propvar);
		ATLASSERT(SUCCEEDED(hResult));
		::PropVariantClear(&propvar);

		hResult = ::InitPropVariantFromString(L"CREATOR", &propvar);
		hResult = ::PSCoerceToCanonicalValue(PKEY_ApplicationName, &propvar);
		hResult = spPropertyStore->SetValue(PKEY_ApplicationName, propvar);
		ATLASSERT(SUCCEEDED(hResult));
		::PropVariantClear(&propvar);

		hResult = spPropertyStore->Commit();
		ATLASSERT(SUCCEEDED(hResult));
		hResult = spPropertyStore->SetValue(PKEY_Null, propvar);
		ATLASSERT(hResult == E_UNEXPECTED);
		spPropertyStore.Release();

		ATLTRACE(TEXT("6:Verify\n"));

		hResult = spClassFactory->CreateInstance(NULL, IID_PPV_ARGS(&spPropertyStore));
		ATLASSERT(SUCCEEDED(hResult));
		spInitializeWithFile = spPropertyStore;
		ATLASSERT(spInitializeWithFile);
		hResult = spInitializeWithFile->Initialize(L"target/normal.pdf", STGM_READ);
		ATLASSERT(SUCCEEDED(hResult));
		spInitializeWithFile.Release();

		hResult = spPropertyStore->GetValue(PKEY_Title, &propvar);
		ATLASSERT(SUCCEEDED(hResult));
		ATLASSERT(propvar.vt == VT_LPWSTR);
		ATLASSERT(wcscmp(propvar.pwszVal, L"TITLE") == 0);
		::PropVariantClear(&propvar);

		hResult = spPropertyStore->GetValue(PKEY_Subject, &propvar);
		ATLASSERT(SUCCEEDED(hResult));
		ATLASSERT(propvar.vt == VT_LPWSTR);
		ATLASSERT(wcscmp(propvar.pwszVal, L"SUBJECT") == 0);
		::PropVariantClear(&propvar);

		hResult = spPropertyStore->GetValue(PKEY_Author, &propvar);
		ATLASSERT(SUCCEEDED(hResult));
		ATLASSERT(propvar.vt == (VT_VECTOR | VT_LPWSTR));
		ATLASSERT(propvar.calpwstr.cElems == 2);
		ATLASSERT(wcscmp(propvar.calpwstr.pElems[0], L"AUTHOR1") == 0);
		ATLASSERT(wcscmp(propvar.calpwstr.pElems[1], L"AUTHOR2") == 0);
		::PropVariantClear(&propvar);

		hResult = spPropertyStore->GetValue(PKEY_Keywords, &propvar);
		ATLASSERT(SUCCEEDED(hResult));
		ATLASSERT(propvar.vt == (VT_VECTOR | VT_LPWSTR));
		ATLASSERT(propvar.calpwstr.cElems == 2);
		ATLASSERT(wcscmp(propvar.calpwstr.pElems[0], L"KEYWORD1") == 0);
		ATLASSERT(wcscmp(propvar.calpwstr.pElems[1], L"KEYWORD2") == 0);
		::PropVariantClear(&propvar);

		hResult = spPropertyStore->GetValue(PKEY_Comment, &propvar);
		ATLASSERT(SUCCEEDED(hResult));
		ATLASSERT(propvar.vt == VT_EMPTY);
		::PropVariantClear(&propvar);

		hResult = spPropertyStore->GetValue(PKEY_Document_TotalEditingTime, &propvar);
		ATLASSERT(SUCCEEDED(hResult));
		ATLASSERT(propvar.vt == VT_EMPTY);
		::PropVariantClear(&propvar);

		hResult = spPropertyStore->GetValue(PKEY_Document_DatePrinted, &propvar);
		ATLASSERT(SUCCEEDED(hResult));
		ATLASSERT(propvar.vt == VT_EMPTY);
		::PropVariantClear(&propvar);

		hResult = spPropertyStore->GetValue(PKEY_Document_DateCreated, &propvar);
		ATLASSERT(SUCCEEDED(hResult));
		ATLASSERT(propvar.vt == VT_FILETIME);
		ATLASSERT(::CompareFileTime(&ftJustNow, &propvar.filetime) == 0);
		::PropVariantClear(&propvar);

		hResult = spPropertyStore->GetValue(PKEY_Document_DateSaved, &propvar);
		ATLASSERT(SUCCEEDED(hResult));
		ATLASSERT(propvar.vt == VT_FILETIME);
		ATLASSERT(::CompareFileTime(&ftJustNow, &propvar.filetime) == 0);
		::PropVariantClear(&propvar);

		hResult = spPropertyStore->GetValue(PKEY_ApplicationName, &propvar);
		ATLASSERT(SUCCEEDED(hResult));
		ATLASSERT(propvar.vt == VT_LPWSTR);
		ATLASSERT(wcscmp(propvar.pwszVal, L"CREATOR") == 0);
		::PropVariantClear(&propvar);

		spPropertyStore.Release();

		::Sleep(300);

		ATLTRACE(TEXT("6:Done\n"));
	}
	#endif

	#if 1 // Test 7: Removing properties
	{
		ATLTRACE(TEXT("7:Start\n"));

		CComPtr<IPropertyStore> spPropertyStore;
		hResult = spClassFactory->CreateInstance(NULL, IID_PPV_ARGS(&spPropertyStore));
		ATLASSERT(SUCCEEDED(hResult));

		CComQIPtr<IInitializeWithFile> spInitializeWithFile = spPropertyStore;
		ATLASSERT(spInitializeWithFile);
		hResult = spInitializeWithFile->Initialize(L"target/normal.pdf", STGM_READWRITE);
		ATLASSERT(SUCCEEDED(hResult));
		spInitializeWithFile.Release();

		hResult = spPropertyStore->SetValue(PKEY_Title, propvar);
		ATLASSERT(SUCCEEDED(hResult));
		hResult = spPropertyStore->SetValue(PKEY_Subject, propvar);
		ATLASSERT(SUCCEEDED(hResult));
		hResult = spPropertyStore->SetValue(PKEY_Author, propvar);
		ATLASSERT(SUCCEEDED(hResult));
		hResult = spPropertyStore->SetValue(PKEY_Keywords, propvar);
		ATLASSERT(SUCCEEDED(hResult));
		hResult = spPropertyStore->SetValue(PKEY_Comment, propvar);
		ATLASSERT(hResult == E_INVALIDARG);
		hResult = spPropertyStore->SetValue(PKEY_Document_TotalEditingTime, propvar);
		ATLASSERT(hResult == E_INVALIDARG);
		hResult = spPropertyStore->SetValue(PKEY_Document_DatePrinted, propvar);
		ATLASSERT(hResult == E_INVALIDARG);
		hResult = spPropertyStore->SetValue(PKEY_Document_DateCreated, propvar);
		ATLASSERT(SUCCEEDED(hResult));
		hResult = spPropertyStore->SetValue(PKEY_Document_DateSaved, propvar);
		ATLASSERT(SUCCEEDED(hResult));
		hResult = spPropertyStore->SetValue(PKEY_ApplicationName, propvar);
		ATLASSERT(SUCCEEDED(hResult));

		hResult = spPropertyStore->Commit();
		ATLASSERT(SUCCEEDED(hResult));
		spPropertyStore.Release();

		ATLTRACE(TEXT("7:Verify\n"));

		hResult = spClassFactory->CreateInstance(NULL, IID_PPV_ARGS(&spPropertyStore));
		ATLASSERT(SUCCEEDED(hResult));
		spInitializeWithFile = spPropertyStore;
		ATLASSERT(spInitializeWithFile);
		hResult = spInitializeWithFile->Initialize(L"target/normal.pdf", STGM_READ);
		ATLASSERT(SUCCEEDED(hResult));
		spInitializeWithFile.Release();

		hResult = spPropertyStore->GetValue(PKEY_Title, &propvar);
		ATLASSERT(SUCCEEDED(hResult));
		ATLASSERT(propvar.vt == VT_EMPTY);
		hResult = spPropertyStore->GetValue(PKEY_Subject, &propvar);
		ATLASSERT(SUCCEEDED(hResult));
		ATLASSERT(propvar.vt == VT_EMPTY);
		hResult = spPropertyStore->GetValue(PKEY_Author, &propvar);
		ATLASSERT(SUCCEEDED(hResult));
		ATLASSERT(propvar.vt == VT_EMPTY);
		hResult = spPropertyStore->GetValue(PKEY_Keywords, &propvar);
		ATLASSERT(SUCCEEDED(hResult));
		ATLASSERT(propvar.vt == VT_EMPTY);
		hResult = spPropertyStore->GetValue(PKEY_Comment, &propvar);
		ATLASSERT(SUCCEEDED(hResult));
		ATLASSERT(propvar.vt == VT_EMPTY);
		hResult = spPropertyStore->GetValue(PKEY_Document_TotalEditingTime, &propvar);
		ATLASSERT(SUCCEEDED(hResult));
		ATLASSERT(propvar.vt == VT_EMPTY);
		hResult = spPropertyStore->GetValue(PKEY_Document_DatePrinted, &propvar);
		ATLASSERT(SUCCEEDED(hResult));
		ATLASSERT(propvar.vt == VT_EMPTY);
		hResult = spPropertyStore->GetValue(PKEY_Document_DateCreated, &propvar);
		ATLASSERT(SUCCEEDED(hResult));
		ATLASSERT(propvar.vt == VT_EMPTY);
		hResult = spPropertyStore->GetValue(PKEY_Document_DateSaved, &propvar);
		ATLASSERT(SUCCEEDED(hResult));
		ATLASSERT(propvar.vt == VT_EMPTY);
		hResult = spPropertyStore->GetValue(PKEY_ApplicationName, &propvar);
		ATLASSERT(SUCCEEDED(hResult));
		ATLASSERT(propvar.vt == VT_EMPTY);

		spPropertyStore.Release();

		::Sleep(300);

		ATLTRACE(TEXT("7:Done\n"));
	}
	#endif

	#if 1 // Test 8: Read-only file
	{
		ATLTRACE(TEXT("8:Start\n"));

		CComPtr<IPropertyStore> spPropertyStore;
		hResult = spClassFactory->CreateInstance(NULL, IID_PPV_ARGS(&spPropertyStore));
		ATLASSERT(SUCCEEDED(hResult));

		CComQIPtr<IInitializeWithFile> spInitializeWithFile = spPropertyStore;
		ATLASSERT(spInitializeWithFile);
		hResult = spInitializeWithFile->Initialize(L"target/read-only.pdf", STGM_READWRITE);
		ATLASSERT(hResult == STG_E_ACCESSDENIED);
		hResult = spInitializeWithFile->Initialize(L"target/read-only.pdf", STGM_READ);
		ATLASSERT(SUCCEEDED(hResult));
		spInitializeWithFile.Release();

		hResult = spPropertyStore->SetValue(PKEY_Null, propvar);
		ATLASSERT(hResult == STG_E_ACCESSDENIED);
		hResult = spPropertyStore->Commit();
		ATLASSERT(hResult == STG_E_ACCESSDENIED);
		spPropertyStore.Release();

		::Sleep(300);

		ATLTRACE(TEXT("8:Done\n"));
	}
	#endif

	#if 1 // Test 9: No enough space left
	{
		ATLTRACE(TEXT("9:Start\n"));

		CComPtr<IPropertyStore> spPropertyStore;
		hResult = spClassFactory->CreateInstance(NULL, IID_PPV_ARGS(&spPropertyStore));
		ATLASSERT(SUCCEEDED(hResult));

		CComQIPtr<IInitializeWithFile> spInitializeWithFile = spPropertyStore;
		ATLASSERT(spInitializeWithFile);
		hResult = spInitializeWithFile->Initialize(L"target/normal.pdf", STGM_READWRITE);
		ATLASSERT(SUCCEEDED(hResult));
		spInitializeWithFile.Release();

		CStringW strTitle(L'*', 2048);

		hResult = ::InitPropVariantFromString(strTitle, &propvar);
		hResult = ::PSCoerceToCanonicalValue(PKEY_Title, &propvar);
		hResult = spPropertyStore->SetValue(PKEY_Title, propvar);
		ATLASSERT(SUCCEEDED(hResult));
		::PropVariantClear(&propvar);

		hResult = spPropertyStore->Commit();
		ATLASSERT(hResult == STG_E_CANTSAVE);
		spPropertyStore.Release();

		::Sleep(300);

		ATLTRACE(TEXT("9:Done\n"));
	}
	#endif

	#if 1 // Teardown
	{
		ATLTRACE(TEXT("Teardown\n"));

		ATLVERIFY(::SetFileAttributes(TEXT("target/normal.pdf"), FILE_ATTRIBUTE_NORMAL));
		ATLVERIFY(::DeleteFile(TEXT("target/normal.pdf")));
		ATLVERIFY(::SetFileAttributes(TEXT("target/read-only.pdf"), FILE_ATTRIBUTE_NORMAL));
		ATLVERIFY(::DeleteFile(TEXT("target/read-only.pdf")));
		ATLVERIFY(::SetFileAttributes(TEXT("target/empty.txt"), FILE_ATTRIBUTE_NORMAL));
		ATLVERIFY(::DeleteFile(TEXT("target/empty.txt")));
	}
	#endif

	::CoUninitialize();
}
